#include <windows.h>
#include "../source/version.h"

#define APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// Version
/////////////////////////////////////////////////////////////////////////////
VS_VERSION_INFO VERSIONINFO
 FILEVERSION MAJOR_VERSION_INT,SUB_VERSION_INT,RELEASE_NUMBER_INT,BUILD_NUMBER_INT
 PRODUCTVERSION MAJOR_VERSION_INT,SUB_VERSION_INT,RELEASE_NUMBER_INT,BUILD_NUMBER_INT
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x40004L
 FILETYPE 0x1L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040004e4"
        BEGIN
            VALUE "FileVersion",      FULL_VERSION_STR"\0"
            VALUE "ProductVersion",   FULL_VERSION_STR"\0"
            VALUE "OriginalFilename", stringOriginalFilename"\0"
            VALUE "FileDescription",  stringFileDescription"\0"
            VALUE "InternalName",     stringFileDescription"\0"
            VALUE "ProductName",      stringFileDescription"\0"
            VALUE "CompanyName",      stringCompanyName"\0"
            VALUE "LegalCopyright",   stringLegalCopyright"\0"
            VALUE "LegalTrademarks",  stringLegalTrademarks"\0"
            //VALUE "PrivateBuild",     " \0"
            //VALUE "SpecialBuild",     " \0"
            //VALUE "Comments",         " \0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x400, 1252
    END
END
